@@ -6,7 +6,7 @@ JSON::Validator - Validate data against a JSON schema
 
 =head1 VERSION
 
-0.51
+0.52
 
 =head1 SYNOPSIS
 
@@ -41,7 +41,7 @@ validate data against L<draft 4|https://github.com/json-schema/json-schema/tree/
 of the specification.
 
 This module is currently EXPERIMENTAL. Hopefully nothing drastic will change,
-but it need to fit together nicely with L<Swagger2> - Since this is a spin-off
+but it needs to fit together nicely with L<Swagger2> - Since this is a spin-off
 project.
 
 =head2 Supported schema formats
@@ -56,7 +56,7 @@ performs the best, so it might change in the future.
 
 =head2 Resources
 
-Here are some resources that is related to JSON schemas and validation:
+Here are some resources that are related to JSON schemas and validation:
 
 =over 4
 
@@ -75,6 +75,7 @@ Here are some resources that is related to JSON schemas and validation:
 =cut
 
 use Mojo::Base -base;
+use Exporter 'import';
 use Mojo::JSON;
 use Mojo::JSON::Pointer;
 use Mojo::URL;
@@ -92,11 +93,29 @@ use constant DEBUG => $ENV{JSON_VALIDATOR_DEBUG} || $ENV{SWAGGER2_DEBUG} || 0;
 use constant WARN_ON_MISSING_FORMAT => $ENV{JSON_VALIDATOR_WARN_ON_MISSING_FORMAT}
   || $ENV{SWAGGER2_WARN_ON_MISSING_FORMAT} ? 1 : 0;
 
-our $VERSION = '0.51';
+our $VERSION = '0.52';
+our @EXPORT_OK = qw( validate_json );
 
 sub E { bless {path => $_[0] || '/', message => $_[1]}, 'JSON::Validator::Error'; }
 sub S { Mojo::Util::md5_sum(Data::Dumper->new([@_])->Sortkeys(1)->Useqq(1)->Dump); }
 
+=head1 FUNCTIONS
+
+=head2 validate_json
+
+  use JSON::Validator "validate_json";
+  @errors = validate_json $data, $schema;
+
+This can be useful in web applications:
+
+  @errors = validate_json $c->req->json, "data://main/spec.json";
+
+=cut
+
+sub validate_json {
+  __PACKAGE__->singleton->schema($_[1])->validate($_[0]);
+}
+
 =head1 ATTRIBUTES
 
 =head2 cache_dir
@@ -105,7 +124,7 @@ sub S { Mojo::Util::md5_sum(Data::Dumper->new([@_])->Sortkeys(1)->Useqq(1)->Dump
   $path = $self->cache_dir;
 
 Path to where downloaded spec files should be cached. Defaults to
-C<JSON_VALIDATOR_CACHE_DIR> or the bundled spec files that is shipped
+C<JSON_VALIDATOR_CACHE_DIR> or the bundled spec files that are shipped
 with this distribution.
 
 =head2 coerce
@@ -238,12 +257,12 @@ has ua => sub {
   $self = $self->schema($url);
   $schema = $self->schema;
 
-Used to set a schema from either an data structure or from a URL.
+Used to set a schema from either a data structure or a URL.
 
 C<$schema> will be a L<Mojo::JSON::Pointer> object when loaded,
 and C<undef> by default.
 
-The C<$url> can take many forms, but need to point to a text file in the
+The C<$url> can take many forms, but needs to point to a text file in the
 JSON or YAML format.
 
 =over 4
@@ -282,15 +301,25 @@ sub schema {
   $self;
 }
 
+=head2 singleton
+
+  $self = $class->singleton;
+
+Returns the L<JSON::Validator> object used by L</validate_json>.
+
+=cut
+
+sub singleton { state $validator = shift->new }
+
 =head2 validate
 
   @errors = $self->validate($data);
 
 Validates C<$data> against a given JSON L</schema>. C<@errors> will
-contain objects with containing the validation errors. It will be
+contain validation error objects. It will be
 empty on success.
 
-Example error element:
+Example error object:
 
   bless {
     message => "Some description",
@@ -425,21 +454,21 @@ sub _resolve_schema {
 
 sub _validate {
   my ($self, $data, $path, $schema) = @_;
-  my ($type) = (map { $schema->{$_} } grep { $schema->{$_} } qw( type allOf anyOf oneOf ))[0];
   my $check_all = grep { $schema->{$_} } qw( allOf oneOf );
-  my @errors;
+  my (@errors, @types);
 
-  $type = 'object' if !$type and $schema->{properties};
-  $type ||= 'any';
+  push @types, map { $schema->{$_} } grep { $schema->{$_} } qw( allOf anyOf oneOf );
+  push @types, $schema->{type} || _guess_schema_type($schema) || 'any' unless @types;
 
+  #warn Carp::longmess(Data::Dumper::Dumper([$data, $path, $schema]));
   #$SIG{__WARN__} = sub { Carp::confess(Data::Dumper::Dumper($schema)) };
 
   if ($schema->{not}) {
     @errors = $self->_validate($data, $path, $schema->{not});
-    return @errors ? () : (E $path, "Should not match.");
+    return @errors ? () : (E $path, 'Should not match.');
   }
 
-  for my $t (ref $type eq 'ARRAY' ? @$type : ($type)) {
+  for my $t (map { ref $_ eq 'ARRAY' ? @$_ : $_ } @types) {
     $t //= 'null';
     if (ref $t eq 'HASH') {
       push @errors, [$self->_validate($data, $path, $t)];
@@ -497,7 +526,7 @@ sub _validate_additional_properties {
   my @errors;
 
   if (ref $properties eq 'HASH') {
-    push @errors, $self->_validate_properties($data, $path, $schema);
+    push @errors, $self->_validate($_, $path, $properties) for values %$data;
   }
   elsif (!$properties) {
     my @keys = grep { $_ !~ /^(description|id|title)$/ } keys %$data;
@@ -510,7 +539,7 @@ sub _validate_additional_properties {
   return @errors;
 }
 
-sub _validate_enum {
+sub _validate_type_enum {
   my ($self, $data, $path, $schema) = @_;
   my $enum = $schema->{enum};
   my $m    = S $data;
@@ -566,7 +595,7 @@ sub _validate_properties {
     my $p = $properties->{$name};
     if (exists $data->{$name}) {
       my $v = delete $data->{$name};
-      push @errors, $self->_validate_enum($v, $path, $p) if $p->{enum};
+      push @errors, $self->_validate_type_enum($v, $path, $p) if $p->{enum};
       push @errors, $self->_validate($v, _path($path, $name), $p);
     }
     elsif ($p->{default}) {
@@ -615,11 +644,11 @@ sub _validate_type_array {
     }
   }
   if (ref $schema->{items} eq 'ARRAY') {
-    my $additional_items = $schema->{additionalItems} // 1;
+    my $additional_items = $schema->{additionalItems} // {type => 'any'};
     my @v = @{$schema->{items}};
 
     if ($additional_items) {
-      push @v, $a while @v < @$data;
+      push @v, $additional_items while @v < @$data;
     }
 
     if (@v == @$data) {
@@ -649,7 +678,7 @@ sub _validate_type_array {
 sub _validate_type_boolean {
   my ($self, $value, $path, $schema) = @_;
 
-  return if defined $value and ("$value" eq "1" or "$value" eq "0");
+  return if defined $value and Scalar::Util::blessed($value) and ("$value" eq "1" or "$value" eq "0");
   return E $path, _expected(boolean => $value);
 }
 
@@ -775,12 +804,12 @@ sub _cmp {
 }
 
 sub _expected {
-  my $type = _guess($_[1]);
+  my $type = _guess_data_type($_[1]);
   return "Expected $_[0] - got different $type." if $_[0] =~ /\b$type\b/;
   return "Expected $_[0] - got $type.";
 }
 
-sub _guess {
+sub _guess_data_type {
   local $_ = $_[0];
   my $ref     = ref;
   my $blessed = Scalar::Util::blessed($_[0]);
@@ -793,6 +822,23 @@ sub _guess {
   return $blessed || 'string';
 }
 
+sub _guess_schema_type {
+  return 'object' if $_[0]->{additionalProperties};
+  return 'object' if $_[0]->{patternProperties};
+  return 'object' if $_[0]->{properties};
+  return 'object' if defined $_[0]->{maxProperties} or defined $_[0]->{minProperties};
+  return 'array'  if $_[0]->{additionalItems};
+  return 'array'  if $_[0]->{items};
+  return 'array'  if $_[0]->{uniqueItems};
+  return 'array'  if defined $_[0]->{maxItems} or defined $_[0]->{minItems};
+  return 'string' if $_[0]->{pattern};
+  return 'string' if defined $_[0]->{maxLength} or defined $_[0]->{minLength};
+  return 'number' if $_[0]->{multipleOf};
+  return 'number' if defined $_[0]->{maximum} or defined $_[0]->{minimum};
+  return 'enum'   if $_[0]->{enum};
+  return;
+}
+
 sub _is_byte_string { $_[0] =~ /^[A-Za-z0-9\+\/\=]+$/; }
 sub _is_date        { $_[0] =~ qr/^(\d+)-(\d+)-(\d+)$/io; }
 sub _is_date_time   { $_[0] =~ qr/^(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+(?:\.\d+)?)(?:Z|([+-])(\d+):(\d+))?$/io; }
